import java.io.*;
import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

public class avp2wpt {
	private static final String appTitle = "avp2wpt";
	private static final String appVersion = "0.1.1";
	private static final String appDesc = "Makes a gpx waypoints file from audio, video or photographic files, and a gpx trace.";
	static Boolean verbose = false;
	static Element inRoot;
	static Date firstTime, lastTime;
	static int timeShift = 0;
	static int wpts = 0;
	static org.jdom.Document inDocument = new Document(inRoot);
	static org.jdom.Element outRoot = new Element("gpx");
	static org.jdom.Document outDocument = new Document(outRoot);
	static int totalPoints = 0;
	static List<Point> points = new LinkedList<Point>();
	static List<File> files = new LinkedList<File>();

	public static void main(String[] args) {
		if (args.length<2) noArgsAndExit();
		if (helpWanted (args)) helpAndExit ();
		if (!checkFile (args[0])) noFileAndExit(args[0]);
		verbose = checkVerbosity (args);
		checkTimeShift (args);
		readTrace(args[0]);
		convertTrace();
		scanFiles (args[1]);
		if (verbose) displayGpx();
		exportGpx(args);
	}

	private static void checkTimeShift(String[] args) {
		for (String arg : args) {
			if (arg.toLowerCase().startsWith("-s"))
				try {
					timeShift = new Integer (arg.substring(2));
				} catch (NumberFormatException e) {
					System.out.println(arg + " : not a valid integer : O will be used.");
					e.printStackTrace();
			}
		}
		if (verbose) System.out.println("Time shift : " + timeShift);
	}

	private static boolean checkVerbosity(String[] args) {
		for (String arg : args) if (arg.toLowerCase().equals("-v")) return true;	
		return false;
	}

	private static void scanFiles(String ext) {
		System.out.println("Scanning " + ext + " files...");
		File[] allFolder = new File(".").listFiles();
		for (File file : allFolder) {
			if (!file.isFile()) continue; //not a normal file
			if (!file.getName().toLowerCase().endsWith(ext.toLowerCase())) continue; //wrong file type
			Date fileDate = new Date(file.lastModified() + (timeShift * 60000));
			if (fileDate.before(firstTime)) {
				if (verbose) System.out.println(file.getName() + " too old."); 
				continue;
			}
			if (fileDate.after(lastTime)) {
				if (verbose) System.out.println(file.getName() + " too recent."); 
				continue;
			}
			if (verbose) System.out.println(file.getName() + " OK.");
			geoLocate(file, fileDate);
		}
	}

	private static void geoLocate(File file, Date fileDate) {
		for (Point point : points) {
			if (point.time().after(fileDate)) {
				Element waypoint = new Element("wpt");
				outRoot.addContent(waypoint);
				waypoint.setAttribute(new Attribute("lat",point.latitude().toString()));
				waypoint.setAttribute(new Attribute("lon",point.longitude().toString()));
				Element link = new Element("link");
				waypoint.addContent(link);
				link.setAttribute("href",file.getName());
				Element text = new Element("text");
				link.addContent(text);
				text.setText (file.getName());
				wpts++;
				break;
			}
		}
	}

	private static boolean checkFile(String filePath) {
		File file = new File(filePath);
	    return file.exists();
	}

	static void readTrace (String fileName) {
		System.out.println("Reading " + fileName);
		SAXBuilder sxb = new SAXBuilder();
		try {
			inDocument = sxb.build(new File(fileName));
		}
		catch(Exception e){}
		inRoot = inDocument.getRootElement();
	}
	static void convertTrace () {
		System.out.print("Converting" + (verbose ? "\n":""));
		if (verbose) System.out.println(inRoot.getName());
		List<Element> tracks = inRoot.getChildren();
		for (Element track : tracks) {
			if (track.getName() != "trk") continue;
			if (verbose) System.out.println("\t" + track.getName());
			List<Element> segments =track.getChildren();
			for (Element segment : segments) {
				if (segment.getName() != "trkseg") continue;
				if (verbose) System.out.println("\t\t" + segment.getName());
				List<Element> trackPts = segment.getChildren();
				for (Element trackPt : trackPts){
					if (trackPt.getName() != "trkpt") continue;
					totalPoints++;
					double latitude = new Double(trackPt.getAttributeValue("lat"));
					double longitude = new Double(trackPt.getAttributeValue("lon"));
					Date time = new Date();
					if (verbose) {
						System.out.print("\t\t\t" + trackPt.getName() + " " + totalPoints);
						System.out.println(" " + latitude + " ; " + longitude);
					}
					else System.out.print(".");
					List<Element> details = trackPt.getChildren();
					for (Element detail : details) {
						if (detail.getName() == "time") {
							String gpsDatePattern = "yyyy-MM-dd'T'HH:mm:ss";
							SimpleDateFormat  gpsDateFormat = new SimpleDateFormat(gpsDatePattern);
							try {
								time = gpsDateFormat.parse(detail.getText().substring(0, 22) + detail.getText().substring(23));
								if (verbose) System.out.println("\t\t\t\t" + time);
							}
							catch (ParseException e) {
								e.printStackTrace();
							}
						}
					}
					points.add(new Point(latitude, longitude, time));
				}
			}
		}
		firstTime = ((LinkedList<Point>) points).getFirst().time();
		lastTime = ((LinkedList<Point>) points).getLast().time();
		System.out.println(verbose ? "" : "\n" + totalPoints + " points read from file.");
		System.out.println("First point : " + firstTime);
		System.out.println("Last point : " + lastTime);
		}

	static void displayGpx() {
		try {
			XMLOutputter sortie = new XMLOutputter(Format.getPrettyFormat());
			sortie.output(outDocument, System.out);
		}
		catch (java.io.IOException e){}
	}

	static void exportGpx(String[] args) {
		if (wpts==0) System.out.println("No waypoint found ; nothing to do.");
		else {
			String fileName = args[0].substring(0, args[0].length()-4) +"-" + args[1]+".gpx";
			try {
				XMLOutputter sortie = new XMLOutputter(Format.getPrettyFormat());
				sortie.output(outDocument, new FileOutputStream(fileName));
			}
			catch (java.io.IOException e){}
			System.out.println(wpts + " waypoints found.");
			System.out.println(fileName + " written.");
		}
	}
	
	static void displayTitle () {
		System.out.println(appTitle + " " + appVersion + " - " + appDesc);
		System.out.println("Author Gilles BONNARD 2009 - gbonnard@tilt-services.com - Made in France");
		System.out.println("Tilt Services - http://www.tilt-services.com/batchtoys");
		System.out.println("This program is free software under the GNU General Public License.");
		System.out.println("Uses the JDOM library : http://www.jdom.org/\n");
	}

	private static void noArgsAndExit() {
		displayTitle();
		System.out.println("Missing arguments : use -h for help.");
		System.exit(1);
	}

	private static void noFileAndExit(String filePath) {
		System.out.println(filePath + " not found.");
		System.exit(2);
	}
	private static void helpAndExit() {
		displayTitle();
		System.out.println("Usage : avp2wpt <FilePath> <ext> [options...]");
		System.out.println("<FilePath> : Path of gpx trace file to be parsed.");
		System.out.println("<ext> : extention of files to treat as waypoints (ie. avi, wav, jpg...).");
		System.out.println("[options...] :");
		System.out.println("\t -v : verbose");
		System.out.println("\t -s : time shift in minutes");
		System.out.println("\t -h : help");
		System.exit(0);
	}

	private static boolean helpWanted(String[] args) {
		for (String arg : args) if (arg.toLowerCase().equals("-h")) return true;	
		return false;
	}
}
